// includes, system
#include <stdio.h>
#include <conio.h>
#include <chrono>
#include <ctime>
#include <iomanip>

using namespace std;

// includes CUDA Runtime
#include <cuda_runtime.h>

// includes, project
#include <helper_cuda.h>
#include <helper_functions.h> // helper utility functions 

//    
#define CIPHER	0x2c308ffbf5e843ba
#define UC			0x504278

union UINT64_VAL
{
	UINT64 Val;
	DWORD DW[2];
	WORD W[4];
};

union DWORD_VAL
{
	DWORD Val;
	BYTE B[4];
};

union WORD_VAL
{
	WORD Val;
	BYTE B[2];
};

typedef struct
{
	WORD K[16];
	DWORD K89, K1011, K1213, K1415;
	UINT64_VAL Res;

} FEAL;

__device__ int d_someoneFoundIt = 0;
__device__ UINT64 d_Key;
__device__ DWORD d_Plain;
__constant__ DWORD d_SelectCode;
__constant__ UINT64_VAL d_Cipher;
__constant__ BYTE d_ROT2[256] =
{
	0x00,0x04,0x08,0x0c,0x10,0x14,0x18,0x1c,0x20,0x24,0x28,0x2c,0x30,0x34,0x38,0x3c,0x40,0x44,0x48,0x4c,0x50,0x54,0x58,0x5c,0x60,0x64,0x68,0x6c,0x70,0x74,0x78,0x7c,
	0x80,0x84,0x88,0x8c,0x90,0x94,0x98,0x9c,0xa0,0xa4,0xa8,0xac,0xb0,0xb4,0xb8,0xbc,0xc0,0xc4,0xc8,0xcc,0xd0,0xd4,0xd8,0xdc,0xe0,0xe4,0xe8,0xec,0xf0,0xf4,0xf8,0xfc,
	0x01,0x05,0x09,0x0d,0x11,0x15,0x19,0x1d,0x21,0x25,0x29,0x2d,0x31,0x35,0x39,0x3d,0x41,0x45,0x49,0x4d,0x51,0x55,0x59,0x5d,0x61,0x65,0x69,0x6d,0x71,0x75,0x79,0x7d,
	0x81,0x85,0x89,0x8d,0x91,0x95,0x99,0x9d,0xa1,0xa5,0xa9,0xad,0xb1,0xb5,0xb9,0xbd,0xc1,0xc5,0xc9,0xcd,0xd1,0xd5,0xd9,0xdd,0xe1,0xe5,0xe9,0xed,0xf1,0xf5,0xf9,0xfd,
	0x02,0x06,0x0a,0x0e,0x12,0x16,0x1a,0x1e,0x22,0x26,0x2a,0x2e,0x32,0x36,0x3a,0x3e,0x42,0x46,0x4a,0x4e,0x52,0x56,0x5a,0x5e,0x62,0x66,0x6a,0x6e,0x72,0x76,0x7a,0x7e,
	0x82,0x86,0x8a,0x8e,0x92,0x96,0x9a,0x9e,0xa2,0xa6,0xaa,0xae,0xb2,0xb6,0xba,0xbe,0xc2,0xc6,0xca,0xce,0xd2,0xd6,0xda,0xde,0xe2,0xe6,0xea,0xee,0xf2,0xf6,0xfa,0xfe,
	0x03,0x07,0x0b,0x0f,0x13,0x17,0x1b,0x1f,0x23,0x27,0x2b,0x2f,0x33,0x37,0x3b,0x3f,0x43,0x47,0x4b,0x4f,0x53,0x57,0x5b,0x5f,0x63,0x67,0x6b,0x6f,0x73,0x77,0x7b,0x7f,
	0x83,0x87,0x8b,0x8f,0x93,0x97,0x9b,0x9f,0xa3,0xa7,0xab,0xaf,0xb3,0xb7,0xbb,0xbf,0xc3,0xc7,0xcb,0xcf,0xd3,0xd7,0xdb,0xdf,0xe3,0xe7,0xeb,0xef,0xf3,0xf7,0xfb,0xff
};
__constant__ DWORD d_UC;

__device__ __forceinline__ BYTE S0( BYTE X1, BYTE X2 )
{
	return d_ROT2[(X1 + X2) & 0xFF];
}

__device__ __forceinline__ BYTE S1( BYTE X1, BYTE X2 )
{
	return d_ROT2[(X1 + X2 + 1) & 0xFF];
}

// Evaluate the FK function.
//
__device__ __forceinline__ DWORD FK( DWORD AA, DWORD BB )
{
	BYTE FK1, FK2;
	DWORD_VAL A, B, C;

	A.Val = AA;
	B.Val = BB;
	FK1 = A.B[1] ^ A.B[0];
	FK2 = A.B[2] ^ A.B[3];
	FK1 = S1( FK1, FK2 ^ B.B[0] );
	FK2 = S0( FK2, FK1 ^ B.B[1] );
	C.B[1] = FK1;
	C.B[2] = FK2;
	C.B[0] = S0( A.B[0], FK1 ^ B.B[2] );
	C.B[3] = S1( A.B[3], FK2 ^ B.B[3] );

	return C.Val;
}

// KP points to an array of 8 bytes.
//
__device__ __forceinline__ void SetKey( FEAL& feal, UINT64_VAL KP )
{
	DWORD A, B, D, NewB;

	A = KP.DW[0];
	B = KP.DW[1];

	D = 0;

	for( int i=0; i<8; i++ )
	{
		NewB = FK( A, B ^ D );
		D = A;
		A = B;
		B = NewB;

		((DWORD*)feal.K)[i] = B;
	}

	feal.K89 = *(DWORD*)(feal.K + 8);
	feal.K1011 = *(DWORD*)(feal.K + 10);
	feal.K1213 = *(DWORD*)(feal.K + 12);
	feal.K1415 = *(DWORD*)(feal.K + 14);
}

// Evaluate the f function.
//
__device__ __forceinline__ DWORD f( DWORD AA, WORD BB )
{
	BYTE f1, f2;
	DWORD_VAL A, C;
	WORD_VAL B;

	A.Val = AA;
	B.Val = BB;
	f1 = A.B[1] ^ B.B[0] ^ A.B[0];
	f2 = A.B[2] ^ B.B[1] ^ A.B[3];
	f1 = S1( f1, f2 );
	f2 = S0( f2, f1 );
	C.B[1] = f1;
	C.B[2] = f2;
	C.B[0] = S0( A.B[0], f1 );
	C.B[3] = S1( A.B[3], f2 );

	return C.Val;
}

// Decrypt a block, using the last key set.
//
__device__ __forceinline__ void Decrypt( FEAL& feal )
{
	DWORD L, R, NewL;

	R = d_Cipher.DW[0];
	L = d_Cipher.DW[1];
	R ^= feal.K1213;
	L ^= feal.K1415;
	L ^= R;

	for( int r = 7; r >= 0; --r )
	{
		NewL = R ^ f( L, feal.K[r] );
		R = L;
		L = NewL;
	}

	R ^= L;
	R ^= feal.K1011;
	L ^= feal.K89;

	feal.Res.DW[0] = L;
	feal.Res.DW[1] = R;
}

__global__ void BF( UINT64 inc )
{
	UINT64_VAL idx{ blockIdx.x * blockDim.x + threadIdx.x };

	FEAL feal;

	idx.Val <<= 40;
	idx.Val += ( inc << 24 ) + d_UC;

	SetKey( feal, idx );
	Decrypt( feal );
	if( feal.Res.DW[1] == d_SelectCode )
	{
		if( ( feal.Res.DW[0] & 0xFF000000 ) == 0x37000000 )
		{
			d_Key = idx.Val;
			d_Plain = feal.Res.DW[0];
			atomicAdd( &d_someoneFoundIt, 1 );

			return;
		}
	}
}

string GetTime()
{
	char sNow[30];

	std::time_t now = std::chrono::system_clock::to_time_t( std::chrono::system_clock::now() );
	std::strftime( sNow, sizeof( sNow ), "%Y-%m-%d %H:%M:%S", std::localtime( &now ) );

	return string( sNow );
}

int main( int argc, char *argv[] )
{
	cudaError_t cudaStatus;
	UINT64 val;
	DWORD uc;
	int someoneFoundIt;
	string str;
	char sBuf[128];

	auto hConsole = ::GetStdHandle( STD_OUTPUT_HANDLE );

	cout << argv[0] << " - Starting...\n";

	// This will pick the best possible CUDA capable device
	int devID = findCudaDevice( argc, (const char **)argv );

	// get device name
	cudaDeviceProp deviceProps;
	checkCudaErrors( cudaGetDeviceProperties( &deviceProps, devID ) );
	cout << "CUDA device " << deviceProps.name << "\n\n";

	//  __global__  () BF    3-   .
	//
	int n = 0x100 * 0x100 * 0x100;

	// set kernel launch configuration
	dim3 threads = dim3( 512 );
	dim3 blocks = dim3( n / threads.x );

	try
	{
		//     Cipher  UC
		//
		val = CIPHER;
		cudaStatus = cudaMemcpyToSymbol( d_Cipher, &val, sizeof( val ) );
		if( cudaStatus != cudaSuccess )
			throw "cudaMemcpyToSymbol failed !";

		uc = UC;
		cudaStatus = cudaMemcpyToSymbol( d_UC, &uc, sizeof( uc ) );
		if( cudaStatus != cudaSuccess )
			throw "cudaMemcpyToSymbol failed !";

		StopWatchInterface* timer = NULL;
		sdkCreateTimer( &timer );

		cout << "START " << GetTime() << "\n\n";

		int sLen = 0;

		for( UINT64 i=0; i<0x10000; i++ )
		//for( UINT64 i = 0x80A0; i < 0x10000; i++ ) // 
		{
			sdkResetTimer( &timer );
			sdkStartTimer( &timer );
			{
				BF << <blocks, threads >> > (i);
				checkCudaErrors( cudaDeviceSynchronize() );
			}
			sdkStopTimer( &timer );

			if( (i % 0x10) == 0 )
			{
				auto tm = sdkGetTimerValue( &timer );

				sprintf_s( sBuf, sizeof( sBuf ), "Try = XX.XX.XX.%02X.%02X.%02X.%02X.%02X (%lld/sec.) %s",
					(BYTE)(i >> 8), (BYTE)i, uc >> 16, (uc >> 8) & 0xFF, uc & 0xFF, (UINT64)(0x1000000 / (tm / 1000)), GetTime().c_str() );

				cout << string( sLen, '\b' ) << sBuf;
				sLen = lstrlen( sBuf );
			}

			// ,   ?

			// ,    ,          ,
			//         d_someoneFoundIt[256*256*256]
			//
			cudaStatus = cudaMemcpyFromSymbol( &someoneFoundIt, d_someoneFoundIt, sizeof( someoneFoundIt ) );
			if( cudaStatus != cudaSuccess )
				throw "cudaMemcpyFromSymbol failed !";

			if( someoneFoundIt > 1 )
			{
				::SetConsoleTextAttribute( hConsole, 0x0C );
				cout << "\nA few keys were found. It's need to modify the code to catch each of them !\n";
				::SetConsoleTextAttribute( hConsole, 0x07 ); //set back to black background and white text
			}

			if( someoneFoundIt )
			{
				sLen = 0;

				someoneFoundIt = 0;
				cudaStatus = cudaMemcpyToSymbol( d_someoneFoundIt, &someoneFoundIt, sizeof( someoneFoundIt ) );
				if( cudaStatus != cudaSuccess )
					throw "cudaMemcpyToSymbol failed !";

				UINT64 key;
				cudaStatus = cudaMemcpyFromSymbol( &key, d_Key, sizeof( key ) );
				if( cudaStatus != cudaSuccess )
					throw "cudaMemcpyFromSymbol failed !";

				DWORD plain;
				cudaStatus = cudaMemcpyFromSymbol( &plain, d_Plain, sizeof( plain ) );
				if( cudaStatus != cudaSuccess )
					throw "cudaMemcpyFromSymbol failed !";

				sprintf_s( sBuf, sizeof( sBuf ), "\nKey = 0x%016llX, Plain = 0x%08X (%s)\n", key, plain, GetTime().c_str() );

				::SetConsoleTextAttribute( hConsole, 0x0A );
				cout << sBuf;
				::SetConsoleTextAttribute( hConsole, 0x07 ); //set back to black background and white text

				std::fstream file( "bf.key", std::ios_base::out | std::ios_base::app );
				if( !file.bad() )
				{
					file << sBuf;
					file.close();
				}
			}
		}

		cout << "END\n";
	}
	catch( const char* msg )
	{
		::SetConsoleTextAttribute( hConsole, 0x0C );
		cout << endl << msg;
		::SetConsoleTextAttribute( hConsole, 0x07 ); //set back to black background and white text
	}

	cout << "\nPress any key\n";
	_getch();

	return EXIT_SUCCESS;
}
